from typing import Callable, Optional, Iterable
from typing import TypeVar, Generic

K = TypeVar('K')
V = TypeVar('V')
T = TypeVar('T')
U = TypeVar('U')


class List(Generic[T], list[T]):
    def map(self, func: Callable[[T], U]) -> "List[U]":
        return List([func(v) for v in self])

    def flat_map(self, func: Callable[[T], Iterable[U]]) -> "List[U]":
        return List([e for v in self for e in func(v)])

    def filter(self, predicate: Callable[[T], bool]) -> "List[T]":
        return List([v for v in self if predicate(v)])

    def find(self, predicate: Callable[[T], bool]) -> Optional[T]:
        for v in self:
            if predicate(v):
                return v
        return None



from easy_pysy.core import event
from easy_pysy.core.cli import command, run
from easy_pysy.core.environment import env
from easy_pysy.core.event import Event, on, emit
from easy_pysy.core.lifecycle import context, start, stop, shutdown, AppStarting, AppStarted, AppStopping, AppState
from easy_pysy.core.logging import trace, debug, info, success, warning, error, critical, exception, log
from easy_pysy.core.provider import get, provide
from easy_pysy.core.thread import Interval
from easy_pysy.plugins import api
from easy_pysy.plugins.loop import loop, Loop, get_loop
from easy_pysy.utils.common import uuid, IntSequence
from easy_pysy.utils.decorators import require
from easy_pysy.utils.decorators import retry
from easy_pysy.utils.functional.function import bind, bind_all
from easy_pysy.utils.functional.iterable import List
from easy_pysy.utils.functional.dictionary import Dict
from easy_pysy.utils.generators import tri_wave, float_range
from easy_pysy.utils.markdown import read_md_table
from easy_pysy.utils.json import JSONEncoder



import json
from datetime import datetime


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)



from typing import Optional, Dict, Any

import uvicorn
from fastapi import FastAPI
from starlette.testclient import TestClient as StarletteTestClient

import easy_pysy as ez

fast_api = FastAPI()

websocket = fast_api.websocket
get = fast_api.get
put = fast_api.put
post = fast_api.post
delete = fast_api.delete
options = fast_api.options
head = fast_api.head
patch = fast_api.patch
trace = fast_api.trace


@fast_api.on_event("shutdown")
def shutdown_event():
    ez.stop()


@ez.command(name='start-api')
def start_api(host: str = '0.0.0.0', port: int = 5000):
    uvicorn.run(fast_api, host=host, port=port)


class TestClient(StarletteTestClient):
    def __init__(
            self,
            base_url: str = "http://testserver",
            raise_server_exceptions: bool = True,
            root_path: str = "",
            backend: str = "asyncio",
            backend_options: Optional[Dict[str, Any]] = None):
        super().__init__(
            fast_api,
            base_url=base_url,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            backend=backend,
            backend_options=backend_options,
        )



from loguru import logger

trace = logger.trace
debug = logger.debug
info = logger.info
success = logger.success
warning = logger.warning
error = logger.error
critical = logger.critical
exception = logger.exception
log = logger.log




import typer
from easy_pysy.core.lifecycle import start, stop, context, AppState
from easy_pysy.core import logging

main_typer = typer.Typer()
# main_typer.add_typer(run_typer, name="run")

command = main_typer.command


def run(auto_start=True):
    if auto_start and context.state == AppState.STOPPED:
        start()

    try:
        main_typer()
    except:
        logging.exception('Exception while running CLI')
        stop()



import json
from dataclasses import dataclass
from datetime import datetime
from threading import Thread
from typing import Type, Callable, TypeVar

from easy_pysy.core.environment import env
from easy_pysy.core.logging import logger
from easy_pysy.utils.common import require
from easy_pysy.utils.functional.iterable import List
from easy_pysy.utils.inspect import qual_name
from easy_pysy.utils.model import PropertyBaseModel


log_file = env('ez.core.event.log_file')
rotation = env('ez.core.event.log_rotation')
compression = env('ez.core.event.log_compression')
retention = env('ez.core.event.log_retention')
storage = env('ez.core.event.storage', config_type=bool, default=False)


if log_file:
    logger.add(
        log_file,
        rotation=rotation,
        compression=compression,
        retention=retention,
        format="{message}",
        filter=__name__,
        level="DEBUG"
    )


class Event(PropertyBaseModel):
    at = datetime.now()

    @property
    def event_type(self):
        return qual_name(self)

    class Config:
        arbitrary_types_allowed = True


@dataclass
class EventSubscriber:
    callback: Callable[[Event], str]
    event_type: Type[Event]
    asynchronous: bool


events = List[Event]()
subscribers: list[EventSubscriber] = []


def on(*event_types: Type[Event], asynchronous=False):
    # TODO: auto detect async function?
    def decorator(func):
        for event_type in event_types:
            subscriber = EventSubscriber(func, event_type, asynchronous)
            subscribers.append(subscriber)
        return func
    return decorator


def emit(event: Event):
    event_type = type(event)

    if log_file:
        message = json.dumps(event.dict(), default=str)
        logger.info(message)
    if storage:
        events.append(event)

    # Synchronous event
    for subscriber in _get_subscribers(event_type):
        if not subscriber.asynchronous:
            _notify_subscriber(event, subscriber)
        else:
            _async_notify_subscriber(event, subscriber)


def _get_subscribers(event_type: Type[Event]) -> list[EventSubscriber]:
    return [
        subscriber
        for subscriber in subscribers
        if issubclass(event_type, subscriber.event_type)
    ]


def _notify_subscriber(event: Event, subscriber: EventSubscriber):
    subscriber.callback(event)


def _async_notify_subscriber(event: Event, subscriber: EventSubscriber):
    thread = Thread(target=_notify_subscriber, args=(event, subscriber), daemon=True)
    thread.start()


EventType = TypeVar('EventType', bound=Type[Event])


def find_by_type(event_type: EventType) -> List[EventType]:
    require(storage, "Storage is not activated. Activate it with env: ez.core.event.storage=True")
    return events.filter(lambda event: isinstance(event, event_type))
